/*-
 * Copyright (c) 1998 Dag-Erling Codan Smrgrav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id$
 */

#include <sys/param.h>
#include <sys/stat.h>

#include <com_err.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <fetch.h>

#define MINBUFSIZE	4096

/* Option flags */
int	 A_flag;	/*!   -A: do not follow 302 redirects */
int	 a_flag;	/*    -a: auto retry */
int	 B_size;	/*    -B: buffer size */
int	 b_flag;	/*!   -b: workaround TCP bug */
char	*f_filename;	/*!   -f: file to fetch */
int	 H_flag;	/*    -H: use high port */
char	*h_hostname;	/*!   -h: host to fetch from */
int	 l_flag;	/*    -l: link rather than copy file: URLs */
int	 m_flag;	/* -[Mm]: set local timestamp to remote timestamp */
int	 o_flag;	/*    -o: specify output file */
int	 o_directory;	/*        output file is a directory */
char	*o_filename;	/*        name of output file */
int	 o_stdout;	/*        output file is stdout */
int	 p_flag;	/* -[Pp]: use passive FTP */
int	 R_flag;	/*    -R: don't delete partially transferred files */
int	 r_flag;	/*!   -r: restart previously interrupted transfer */
int	 T_secs = -1;	/*    -T: transfer timeout in seconds */
int	 t_flag;	/*!   -t: workaround TCP bug */
int	 v_level = 1;	/*    -v: verbosity level */
int	 v_tty;		/*        stdout is a tty */
int	 w_secs;	/*    -w: retry delay */


int	 ftp_timeout;	/* default timeout for FTP transfers */
int	 http_timeout;	/* default timeout for HTTP transfers */
char	*buf;		/* transfer buffer */


void
sig_handler(int sig)
{
    errx(1, "Transfer timed out");
}

int
fetch(char *URL, char *path)
{
    struct url *url;
    struct url_stat us;
    struct timeval tv[2];
    float delta, bps;
    FILE *f, *of;
    size_t size;
    off_t count;
    char flags[8];
    int ch, n, timeout;

    if ((url = fetchParseURL(URL)) == NULL)
	errx(1, "%s: parse error", URL);

    timeout = 0;
    *flags = 0;
    
    if (v_level > 2)
	strcat(flags, "v");

    /* FTP specific flags */
    if (strcmp(url->scheme, "ftp") == 0) {
	if (p_flag)
	    strcat(flags, "p");
	if (H_flag)
	    strcat(flags, "h");
	timeout = (T_secs == -1) ? ftp_timeout : T_secs;
    }
    
    /* HTTP specific flags */
    if (strcmp(url->scheme, "http") == 0) {
	timeout = (T_secs == -1) ? http_timeout : T_secs;
    }

    /* symlink instead of copy? */
    if (l_flag) {
	if (strcmp(url->scheme, "file") == 0) {
	    if (symlink(url->doc, path) == -1)
		warn("%s", path);
	    return -1;
	} else
	    warnx("%s: -l flag is not applicable to the '%s' scheme",
		  URL, url->scheme);
    }

    /* output to stdout? */
    if (o_stdout) {
	of = stdout;
    } else if ((of = fopen(path, "w")) == NULL) {
	warn("%s", path);
	return -1;
    }

    if (timeout)
	alarm(timeout);
    if (fetchStat(url, &us, flags) == -1)
	us.size = -1;
    if (timeout)
	alarm(timeout);
    if ((f = fetchGet(url, flags)) == NULL) {
	/* libcom_err has already printed an error message */
	return -1;
    }
    if (timeout)
	alarm(0);
    free(url);
    
    fprintf(stderr, "Recieving %s", URL);
    if (us.size != -1)
	fprintf(stderr, " (%lld bytes)", us.size);
    fputc('\n', stderr);
    
    if (v_level)
	gettimeofday(&tv[0], NULL);

    n = 0;
    
    if (us.size == -1) {
	/*	  
	 * We have no idea how much data to expect, so do it byte by
         * byte. This is incredibly inefficient, but there's not much
         * we can do about it... :(	 
	 */
#ifdef STDIO_HACK
	off_t ocount;
	ocount = 0;
#endif
	count = 0;
	while (1) {
	    if (timeout)
		alarm(timeout);
#ifdef STDIO_HACK
	    /*	      
	     * This is a non-portable hack, but it makes things go
	     * faster. Basically, if there is data in the input file's
	     * buffer, write it out; then fall through to the fgetc()
	     * which forces a refill. It saves a memcpy() and reduces
	     * the number of iterations, i.e the number of calls to
	     * alarm(). Empirical evidence shows this can cut user
	     * time by up to 90%. There may be better (even portable)
	     * ways to do this.
	     */
	    if (f->_r && (f->_ub._base == NULL)) {
		if (fwrite(f->_p, f->_r, 1, of) < 1)
		    break;
		count += f->_r;
		f->_p += f->_r;
		f->_r = 0;
	    }
#endif
	    if ((ch = fgetc(f)) == EOF || fputc(ch, of) == EOF)
		break;
	    if (timeout)
		alarm(0);
	    count++;
#ifdef STDIO_HACK
	    if (((count & !0xFFF) > (ocount & !0xFFF)) && v_level > 1 && v_tty) {
		ocount = count;
#else
	    if ((count & 0xFFF) == 0 && v_level > 1 && v_tty) {
#endif
		fprintf(stderr, "\rReceiving %s: %lld bytes", path, count);
		fflush(stderr);
	    }
#if 0
	    } /* make Emacs happy */
#endif
	    n++;
	}
	if (v_level > 1 && v_tty) {
	    fprintf(stderr, "\rReceiving %s: %lld bytes\n", path, count);
	    fflush(stderr);
	}
    } else {
	/* we now exactly how much to transfer, so do it efficiently */
	if (v_level > 1 && v_tty) {
	    fprintf(stderr, "\rReceiving %s (%lld bytes): 0%%", path, us.size);
	    fflush(stderr);
	}
	for (count = 0, size = B_size; count != us.size; n++) {
	    if (us.size - count < B_size)
		size = us.size - count;
	    if (timeout)
		alarm(timeout);
	    if (fread(buf, size, 1, f) != 1 || fwrite(buf, size, 1, of) != 1)
		break;
	    if (timeout)
		alarm(0);
	    count += size;
	    if (v_level > 1 && v_tty) {
		fprintf(stderr, "\rReceiving %s (%lld bytes): %d%%",
			path, us.size, (int)((100.0 * count) / us.size));
		fflush(stderr);
	    }
	}
	if (v_level > 1 && v_tty) {
	    fputc('\n', stderr);
	    fflush(stderr);
	}
    }

    /* check the status of our files */
    if (ferror(f))
	warn("%s", URL);
    if (ferror(of))
	warn("%s", path);
    if (ferror(f) || ferror(of)) {
	fclose(f);
	fclose(of);
	if (!R_flag && !o_stdout)
	    unlink(path);
	return -1;
    }
    
    /* compute and display transfer speed */
    if (v_level) {
	gettimeofday(&tv[1], NULL);
	delta = (tv[1].tv_sec + (tv[1].tv_usec / 1.e6))
	    - (tv[0].tv_sec + (tv[0].tv_usec / 1.e6));
	if (tv[0].tv_usec < 0) {
	    tv[0].tv_usec += 1000000;
	    tv[0].tv_sec--;
	}
	if (v_level > 1)
	    fprintf(stderr, "%d chunks received\n", n);
	fprintf(stderr, "%lld bytes transferred in %.1f seconds ",
		count, delta);
	bps = count / delta;
	if (bps > 1024*1024)
	    fprintf(stderr, "(%.2f MBps)\n", bps / (1024*1024));
	else if (bps > 1024)
	    fprintf(stderr, "(%.2f kBps)\n", bps / 1024);
	else
	    fprintf(stderr, "(%.2f Bps)\n", bps);
    }

    fclose(f);
    fclose(of);
    
    /* Set mtime of local file */
    if (m_flag && us.size != -1 && !o_stdout) {
	tv[0].tv_sec = (long)us.atime;
	tv[1].tv_sec = (long)us.mtime;
	tv[0].tv_usec = tv[1].tv_usec = 0;
	utimes(path, tv);
    }

    return 0;
}

void
usage(void)
{
    fprintf(stderr,
	    "fetch [-AHMPRalmnpqrv]"
	    " [-B bytes] [-T seconds] [-o file] [-w seconds]"
	    " URL ...\n");
}


int
parseint(char *s, int *v)
{
    *v = 0;
    for (*v = 0; *s; s++)
	if (isdigit(*s))
	    *v = *v * 10 + *s - '0';
	else
	    return -1;
    return 0;
}

int
main(int argc, char *argv[])
{
    struct stat sb;
    char *p, *q, *s;
    int c, e, r;

    while ((c = getopt(argc, argv, "AaB:bf:h:lHMmnPpo:qRrT:tvw:")) != EOF)
	switch (c) {
	case 'A':
	    warnx("warning: the -A option is currently unimplemented");
	    A_flag = 1;
	    break;
	case 'a':
	    a_flag = 1;
	    break;
	case 'B':
	    if (parseint(optarg, &B_size) == -1) {
		errx(1, "invalid buffer size");
	    }
	    break;
	case 'b':
	    warnx("warning: the -b option is deprecated");
	    b_flag = 1;
	    break;
	case 'f':
	    errx(1, "this syntax is deprecated.");
	    break;
	case 'H':
	    H_flag = 1;
	    break;
	case 'h':
	    errx(1, "this syntax is deprecated.");
	    break;
	case 'l':
	    l_flag = 1;
	    break;
	case 'o':
	    o_flag = 1;
	    o_filename = optarg;
	    break;
	case 'M':
	case 'm':
	    m_flag = 1;
	    break;
	case 'n':
	    m_flag = 0;
	    break;
	case 'P':
	case 'p':
	    p_flag = 1;
	    break;
	case 'q':
	    v_level = 0;
	    break;
	case 'R':
	    R_flag = 1;
	    break;
	case 'r':
	    warnx("warning: the -r option is currently unimplemented");
	    r_flag = 1;
	    break;
	case 'T':
	    if (parseint(optarg, &T_secs) == -1) {
		errx(1, "invalid timeout");
	    }
	    break;
	case 't':
	    t_flag = 1;
	    warnx("warning: the -b option is deprecated");
	    break;
	case 'v':
	    v_level++;
	    break;
	case 'w':
	    a_flag = 1;
	    if (parseint(optarg, &w_secs) == -1) {
		errx(1, "invalid delay");
	    }
	    break;
	default:
	    usage();
	    exit(EX_USAGE);
	}

    argc -= optind;
    argv += optind;

    if (!argc) {
	usage();
	exit(EX_USAGE);
    }

    /* allocate buffer */
    if (B_size < MINBUFSIZE)
	B_size = MINBUFSIZE;
    if ((buf = malloc(B_size)) == NULL)
	errx(1, strerror(ENOMEM));

    /* timeout handling */
    signal(SIGALRM, sig_handler);
    if ((s = getenv("FTP_TIMEOUT")) != NULL) {
	if (parseint(s, &ftp_timeout) == -1) {
	    warnx("FTP_TIMEOUT is not a positive integer");
	    ftp_timeout = 0;
	}
    }
    if ((s = getenv("HTTP_TIMEOUT")) != NULL) {
	if (parseint(s, &http_timeout) == -1) {
	    warnx("HTTP_TIMEOUT is not a positive integer");
	    http_timeout = 0;
	}
    }

    /* passive mode */
    if ((s = getenv("FTP_PASSIVE_MODE")) != NULL) {
	if (strcasecmp(s, "yes") == 0)
	    p_flag = 1;
    }
    
    /* output file */
    if (o_flag) {
	if (strcmp(o_filename, "-") == 0) {
	    o_stdout = 1;
	} else if (stat(o_filename, &sb) == -1) {
	    if (errno == ENOENT) {
		if (argc > 1)
		    errx(EX_USAGE, "%s is not a directory", o_filename);
	    } else {
		err(EX_IOERR, "%s", o_filename);
	    }
	} else {
	    if (sb.st_mode & S_IFDIR)
		o_directory = 1;
	}
    }

    /* check if output is to a tty (for progress report) */
    v_tty = isatty(STDOUT_FILENO);
    r = 0;

    while (argc) {
	if ((p = strrchr(*argv, '/')) == NULL)
	    p = *argv;
	else
	    p++;

	fetchLastErrCode = 0;
	
	if (o_flag) {
	    if (o_stdout) {
		e = fetch(*argv, NULL);
	    } else if (o_directory) {
		asprintf(&q, "%s/%s", o_filename, p);
		e = fetch(*argv, q);
		free(q);
	    } else {
		e = fetch(*argv, o_filename);
	    }
	} else {
	    e = fetch(*argv, p);
	}

	if (e) {
	    r = 1;
	    if ((fetchLastErrCode
		 && fetchLastErrCode != FETCH_UNAVAIL
		 && fetchLastErrCode != FETCH_MOVED
		 && fetchLastErrCode != FETCH_URL
		 && fetchLastErrCode != FETCH_RESOLV
		 && fetchLastErrCode != FETCH_UNKNOWN)) {
		if (w_secs) {
		    if (v_level)
			fprintf(stderr, "Waiting %d seconds before retrying", w_secs);
		    sleep(w_secs);
		}
		if (a_flag)
		    continue;
		fprintf(stderr, "Skipping %s", *argv);
	    }
	}

	argc--, argv++;
    }
    
    exit(r);
}
